//
//  APIRequestManager.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import Foundation
import Firebase

/**
 
 Firestore structure:
 
 COLLECTION > DOCUMENT    > COLLECTION > DOCUMENT
 Users      > (user uid)  > (user data)
                            events     > (activity uid) > event data
                                
 Cities     > (city name) > activities > (activity data)
                            places     > (place data)
 
 */

protocol FirebaseFirestoreManager {
    
    //Will fetch a document from the base level collections
    //Like an user from users
    func fetchDocument(
        mainCollection: String,
        document: String,
        completion: @escaping (DocumentSnapshot?, Error?) -> Void)
    
    //For posting data to a top collection
    //Is used to send the user data to the collection
    func postDataWithDoc(
        collection: String,
        document: String,
        data: [String : Any],
        completion: @escaping (_ isSuccessful: Bool) -> Void)
    
    //For posting data to a nested collection (second level)
    //Used to save data of user events, places or activities
    func postDataDeeper(
        mainCollection: String,
        document: String,
        secondaryCollection: String,
        data: [String : Any],
        completion: @escaping (_ isSuccessful: Bool, _ ref: String) -> Void)
    
    //For posting data to a nested collection (second level) with a document id
    //Used to save data of user events, places or activities
    func postDataDeeperWithUID(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        documentUID: String,
        data: [String : Any],
        completion: @escaping (_ isSuccessful: Bool) -> Void)
    
    //Will delete a document from a secondary level conllection
    //For example an event inside user events or an activity
    func deleteFromDeeper(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        document: String,
        completion: @escaping (_ didDelete: Bool)->Void)
    
    func updateWithId(
        collection: String,
        document: String,
        data: [String: Any],
        completion: @escaping (_ isSuccessful: Bool) -> Void)

    //For updating a document in a second level collection
    //such as events of users, or activies / places to cities
    func updateDeeper(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        secondaryDocument: String,
        data: [String: Any],
        completion: @escaping (_ isSuccessful: Bool) -> Void)
    
    //For fetching a nested collection in a document
    //Is used to add events to users, or activies / places to cities
    func fetchDeeper(
        mainCollection: String,
        document: String,
        secondaryCollection: String,
        completion: @escaping (_ data: [QueryDocumentSnapshot], _ error: Error?) -> Void)
    
    //For searching an item in a nested collection
    //Is used to search items in activities or cities
    func searchDeeper(
        query: String,
        field: String,
        mainCollection: String,
        document: String,
        secondaryCollection: String,
        completion: @escaping (_ data: [QueryDocumentSnapshot], _ error: Error?) -> Void)
    
    //For getting a document in collection
    //Is used to get user
    func fetchDocumentInCollection(
        mainCollection: String,
        document: String,
        completion: @escaping (_ data: DocumentSnapshot?, _ error: Error?) -> Void)
    
    //For getting a document in a second level collection
    //Will receive a closure
    func fetchDocumentDeeper(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        document: String,
        completion: @escaping (DocumentSnapshot?, Error?) -> Void)
    
}
