//
//  AddPlaceViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez on 1/12/21.
//

import UIKit
import MapKit
import MobileCoreServices

//Protocol to route back to root
protocol AddPlaceViewRouteProtocol: AnyObject {
    func displayRouteBack()
    func displaySetupView(viewModel: AddPlaceBusinessModel.SetupView.ViewModel)
    func displayResultAlert(text: String)
    func displayLocationInMap(location: Location)
}

//Protocol to display loading
protocol AddPlaceViewLoadingProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
}

class AddPlaceViewController:
    UIViewController,
    UINavigationControllerDelegate,
    UIImagePickerControllerDelegate,
    AddPlaceViewRouteProtocol,
    AddPlaceViewLoadingProtocol {
    
    @IBOutlet weak var validationText: UILabel!
    @IBOutlet weak var map: MKMapView!
    @IBOutlet weak var nameInput: UITextField!
    @IBOutlet weak var descriptionInput: UITextField!
    @IBOutlet weak var tagsInput: UITextField!
    @IBOutlet weak var SaveBtn: UIButton!
    @IBOutlet weak var takePhoto: UIButton!
    @IBOutlet weak var photoView: UIImageView!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    
    var viewModel: AddPlaceBusinessModel.SetupView.ViewModel?
    var imageURL: URL? = nil
    
    let interactor: AddPlaceInteractorProtocol
    let router: AddPlaceRouterProtocol
    
    init(interactor: AddPlaceInteractorProtocol, router: AddPlaceRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: AddPlaceViewController.self)
        super.init(nibName: "\(AddPlaceViewController.self)", bundle: bundle)
    }
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    
    override func viewDidLoad() {
        super.viewDidLoad()
        interactor.viewDidLoad()
        interactor.getUserLocation()
        self.title = "Añadir este Lugar"
    }
    @IBAction func saveClick(_ sender: Any) {
        guard let validator = viewModel?.validator(
                nameInput.text,
                descriptionInput.text,
                tagsInput.text!,
                imageURL) else{
            return
        }
        validationText.isHidden = validator.isValid
        validationText.text = validator.errorMessage
        
        if validator.isValid {
            interactor.savePlace(
                request: AddPlaceBusinessModel.SavePlace.Request(
                    name: validator.nameValue,
                    description: validator.descriptionValue,
                    tags: validator.tagsValue,
                    image: validator.imageValue
                )
            )
        }
    }
    @IBAction func takePhotoClick(_ sender: Any) {
        let vc = UIImagePickerController()
        vc.allowsEditing = true
        vc.delegate = self
        if UIImagePickerController.isSourceTypeAvailable(.camera) {
            vc.sourceType = .camera
        }else{
            vc.sourceType = .photoLibrary
        }
        present(vc, animated: true)
    }
    
    func imagePickerController(_ picker: UIImagePickerController, didFinishPickingMediaWithInfo info: [UIImagePickerController.InfoKey : Any]) {
        picker.dismiss(animated: true)

        guard let image = info[.editedImage] as? UIImage else {
            print("No image found")
            return
        }
        photoView.image = image
        
        // Check for the media type
        let mediaType = info[UIImagePickerController.InfoKey.mediaType] as! CFString
        if mediaType == kUTTypeImage {
            imageURL = info[UIImagePickerController.InfoKey.imageURL] as? URL
        }
        
    }
    
    func displayRouteBack() {
        router.routeBack()
    }
    
    func displaySetupView(viewModel: AddPlaceBusinessModel.SetupView.ViewModel) {
        self.viewModel = viewModel
        SaveBtn.layer.cornerRadius = viewModel.save.cornerRadius
        SaveBtn.layer.borderWidth = viewModel.save.borderWidth
        SaveBtn.layer.borderColor = viewModel.save.borderColor
        //Looks for single or multiple taps.
        let tap = UITapGestureRecognizer(
            target: self,
            action: #selector(UIInputViewController.dismissKeyboard)
        )
        view.addGestureRecognizer(tap)
    }
    
    //Calls this function when the tap is recognized.
    @objc func dismissKeyboard() {
        //Causes the view (or one of its embedded text fields) to resign the first responder status.
        view.endEditing(true)
    }
    
    func displayResultAlert(text: String) {
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Entendido", style: .default, handler: { [weak self] action in
            self?.displayRouteBack()
        }))
        self.present(alert, animated: true, completion: nil)
    }
    
    func displayLoading() {
        loader.isHidden = false
        SaveBtn.isHidden = true
        takePhoto.isEnabled = false
        validationText.isHidden = true
    }
    
    func hideLoading() {
        loader.isHidden = true
        SaveBtn.isHidden = false
        takePhoto.isEnabled = true
        
    }
    
    func displayLocationInMap(location: Location) {
        print("LOCATION: \(location)")
        map.showsUserLocation = true
        let cloc = CLLocationCoordinate2D(latitude: location.latitude, longitude: location.longitude)
        let coordinateRegion = MKCoordinateRegion.init(center: cloc,
                                                       latitudinalMeters: 1000,
                                                       longitudinalMeters: 1000)
        map.setRegion(coordinateRegion, animated: true)
    }
 
    
}

