//
//  AddPlaceWorker.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import Foundation

//Abstranction of worker
//Used by the interactor
protocol AddPlaceWorkerProtocol {
    func savePlace(
        name: String,
        description: String,
        tags: String,
        image: String,
        completion: @escaping (_ didSave: Bool)-> Void)
    func savePlaceImage(
        path: URL,
        name: String,
        completion: @escaping (_ url: String?, _ error: Error?)-> Void)
    func getUserLocation(
        success: @escaping (_ location:Location)->Void,
        failure: @escaping (_ error: Error?)-> Void)
}
//Worker class is in charge of api requests
//Need services injection for usage
//This worker will save the place into the db
class AddPlaceWorker: AddPlaceWorkerProtocol {
    
    let firestore: FirebaseFirestoreManager
    let location: LocationManager
    let firebaseImageUpload: FirebaseImageUploadManager
    
    init(firestore: FirebaseFirestoreManager, location: LocationManager, firebaseImageUpload: FirebaseImageUploadManager){
        self.firestore = firestore
        self.location = location
        self.firebaseImageUpload = firebaseImageUpload
    }
    
    //Function to save place into db
    func savePlace(
        name: String,
        description: String,
        tags: String,
        image: String,
        completion: @escaping (Bool) -> Void) {

        //to save place first needs to check city
        //get user current position
        location.fetchLocation(){ [weak self] position, error in
            if(error != nil){
                completion(false)
                return
            }
            //get city from reverse geolocation
            self?.location.getCityFromLocation(location: position!){ [weak self] city in
                
                if let city = city {
                    //Send data to firestore
                    let data: [String: Any] = [
                        "description":description,
                        "image":image,
                        "lat": position!.latitude,
                        "lng": position!.longitude,
                        "name": name,
                        "tags": tags
                    ]
                    
                    self?.firestore.postDataDeeper(
                        mainCollection: "Cities",
                        document: city,
                        secondaryCollection: "Places",
                        data: data){ didSave, ref in
                        completion(didSave)
                    }
                }else{
                    completion(false)
                }
            }
        }
    }
    
    //This function will call the api that saves images
    //returns true or false
    func savePlaceImage(
        path: URL,
        name: String,
        completion: @escaping (String?, Error?)-> Void){
        firebaseImageUpload.uploadImage(path: path, name: name){ url, error in
            completion(url, error)
        }
    }
    
    func getUserLocation(
        success: @escaping (_ location:Location)->Void,
        failure: @escaping (_ error: Error?)-> Void){
        
        location.fetchLocation(){ location, error in
            if error != nil {
                failure(error!)
                return
            }
            success(location!)
        }
    }
    
}
