//
//  DetailViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez on 1/12/21.
//

import UIKit
import MapKit

//Protocol used by the presenter to pass data
protocol DetailViewControllerProtocol: AnyObject {
    func displayData(viewModel: DetailBusinessModel.LoadData.ViewModel)
    func displaySetupView(viewModel: DetailBusinessModel.SetupView.ViewModel)
    func displayResultAlert(text: String)
    func displayChangeAssistButton(userIsAssisting: Bool)
}
//Protocol to display loading
protocol DetailViewLoadingProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
}
//View controller for item details
//will show more info about places or activities
//Has map to show location
//if is a place will show different data from activity
class DetailViewController: UIViewController, DetailViewControllerProtocol, DetailViewLoadingProtocol {

    
    //View components
    @IBOutlet weak var titleText: UILabel!
    @IBOutlet weak var photoTitle: UILabel!
    @IBOutlet weak var map: MKMapView!
    @IBOutlet weak var dateText: UILabel!
    @IBOutlet weak var attendeesText: UILabel!
    @IBOutlet weak var descriptionText: UILabel!
    @IBOutlet weak var photoView: UIImageView!
    @IBOutlet weak var shareBtn: UIButton!
    @IBOutlet weak var assistBtn: UIButton!
    @IBOutlet weak var newEventBtn: UIButton!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    @IBOutlet weak var whereText: UILabel!
    
    var viewModel: DetailBusinessModel.SetupView.ViewModel?
    var isPlace = false
    var willAssist = false
    
    let interactor: DetailInteractorProtocol
    let router: DetailRouterProtocol
    
    var shareItem: [String] = []
    
    init(interactor: DetailInteractorProtocol, router: DetailRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: DetailViewController.self)
        super.init(nibName: "\(DetailViewController.self)", bundle: bundle)
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        self.title = "Detalle"
        interactor.loadData()
        interactor.viewDidLoad()
    }
    
    //Will show share sheet to send some data
    @IBAction func shareAction(_ sender: Any) {
        let items = shareItem
        let ac = UIActivityViewController(activityItems: items, applicationActivities: nil)
        present(ac, animated: true)
    }
    //Will add one to the counter and send to fb
    @IBAction func assistAction(_ sender: Any) {
        //Tell the interactor to update and post data
        //the interactor will tell the presenter to show a dialog confirming or denying
        //the presenter will tell the view to change the button text
        if(!willAssist){
            willAssist = true
        }else{
            willAssist = false
        }
        interactor.updateAttendees(willAssist: willAssist)
        interactor.addEventToUser(reques: .init(isCreator: false, isGoing: true))
    }
    //will show new event view
    @IBAction func newEventAction(_ sender: Any) {
        router.routeNewEvent()
    }
    
    func displayData(viewModel: DetailBusinessModel.LoadData.ViewModel) {
        
        let cloc = CLLocationCoordinate2D(latitude: viewModel.lat, longitude: viewModel.lng)
        let coordinateRegion = MKCoordinateRegion.init(center: cloc,
                                                       latitudinalMeters: 500,
                                                       longitudinalMeters: 500)
        map.setRegion(coordinateRegion, animated: true)
        let annotation = MKPointAnnotation()
        annotation.coordinate = cloc
        map.addAnnotation(annotation)
        isPlace = viewModel.isPlace
        
        if viewModel.isPlace {
            dateText.isHidden = true
            attendeesText.isHidden = true
            assistBtn.isHidden = true
            newEventBtn.isHidden = false
            photoTitle.isHidden = false
            photoView.isHidden = false
            loader.isHidden = true
            whereText.isHidden = true
            
            titleText.text = viewModel.name
            descriptionText.text = viewModel.description
            viewModel.loadImage(photoView)
            
            let googleLocation = "http://www.google.com/maps/place/\(viewModel.lat),\(viewModel.lng)"
            shareItem =  ["Descubrí este lugar en Quillair App: \(viewModel.name). \(googleLocation)"]
            
        }else{
            dateText.isHidden = false
            attendeesText.isHidden = false
            assistBtn.isHidden = false
            newEventBtn.isHidden = true
            photoTitle.isHidden = true
            photoView.isHidden = true
            loader.isHidden = true
            whereText.isHidden = false
            
            titleText.text = viewModel.name
            descriptionText.text = viewModel.description
            dateText.text = viewModel.date
            attendeesText.text = "Asistentes: \(viewModel.attendees ?? 1)"
            whereText.text = "En \(viewModel.placeName ?? "un lugar")"
            
            
            let googleLocation = "http://www.google.com/maps/place/\(viewModel.lat),\(viewModel.lng)"
            shareItem =  ["Descarga Quillair App para descubrir este evento: \(viewModel.name) en \(viewModel.placeName ?? "este sitio") en la fecha \(viewModel.date ?? "") \(googleLocation)"]
            
        }
    }
    
    func displaySetupView(viewModel: DetailBusinessModel.SetupView.ViewModel) {
        self.viewModel = viewModel
        shareBtn.layer.cornerRadius = viewModel.grayButton.cornerRadius
        shareBtn.layer.borderWidth = viewModel.grayButton.borderWidth
        shareBtn.layer.borderColor = viewModel.grayButton.borderColor
        assistBtn.layer.cornerRadius = viewModel.greenButton.cornerRadius
        assistBtn.layer.borderWidth = viewModel.greenButton.borderWidth
        assistBtn.layer.borderColor = viewModel.greenButton.borderColor
        newEventBtn.layer.cornerRadius = viewModel.greenButton.cornerRadius
        newEventBtn.layer.borderWidth = viewModel.greenButton.borderWidth
        newEventBtn.layer.borderColor = viewModel.greenButton.borderColor
    }
    
    func displayResultAlert(text: String) {
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Entendido", style: .default, handler: { [weak self] action in
            self?.displayRouteBack()
        }))
        self.present(alert, animated: true, completion: nil)
    }
    
    func displayRouteBack() {
        router.routeBack()
    }
    
    func displayLoading() {
        loader.isHidden = false
        assistBtn.isHidden = true
        newEventBtn.isHidden = true
        shareBtn.isHidden = true
    }
    
    func hideLoading() {
        loader.isHidden = true
        shouldShowViewComponents()
    }
    
    private func shouldShowViewComponents(){
        if isPlace {
            dateText.isHidden = true
            attendeesText.isHidden = true
            assistBtn.isHidden = true
            newEventBtn.isHidden = false
            photoTitle.isHidden = false
            photoView.isHidden = false
            loader.isHidden = true
            whereText.isHidden = true
        }else{
            dateText.isHidden = false
            attendeesText.isHidden = false
            assistBtn.isHidden = false
            newEventBtn.isHidden = true
            photoTitle.isHidden = true
            photoView.isHidden = true
            loader.isHidden = true
            whereText.isHidden = false
        }
    }
    
    //If user said they were assisting should hide the button
    //else should show it
    func displayChangeAssistButton(userIsAssisting: Bool){
        assistBtn.isHidden = userIsAssisting
    }
    
}
