//
//  HomeInteractor.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import Foundation

// Abstraction of interactor
protocol HomeInteractorProtocol {
    func getPlaces()
    func getActivities()
    func searchPlaces(query: String)
    func searchActivities(query: String)
    func selectPlace(request: HomeBusinessModel.SelectPlace.Request)
    func selectActivity(request: HomeBusinessModel.SelectActivity.Request)
    func deselect()
}

// Helper to send data to other scenes
protocol HomeDataStore: AnyObject {
    var selectedPlace: Place? { get set }
    var selectedActivity: Activity? { get set }
}

// Is in charge of telling the worker to request data
// passes the data to the presenter
class HomeInteractor: HomeInteractorProtocol, HomeDataStore {

    var selectedPlace: Place?
    var selectedActivity: Activity?
    
    var activities: [Activity] = []
    var places: [Place] = []
    
    let worker: HomeWorkerProtocol
    let presenter: HomePresenterProtocol & HomePresentAlertsProtocol
    
    init(worker: HomeWorkerProtocol, presenter: HomePresenterProtocol & HomePresentAlertsProtocol) {
        self.worker = worker
        self.presenter = presenter
    }
    
    // gets the list of all places
    // will send it to the presenter
    func getPlaces() {
        presenter.presentLoading()
        worker.getPlaces(success: { [weak self] data in
            print(data)
            self?.presenter.removeLoading()
            self?.places = data
            self?.presenter.presentLoadPlaces(response: HomeBusinessModel.LoadData.PlacesResponse(items: data))
        }, failure: { [weak self] error in
            print(error!)
            self?.presenter.removeLoading()
            self?.presenter.presentError(text: "Ha ocurrido un error al cargar los datos.")
        })
    }
    
    // gets the list of all activities
    // will send data to presenter
    func getActivities() {
        presenter.presentLoading()
        worker.getActivities(success: { [weak self] data in
            print("Interactor activities")
            self?.presenter.removeLoading()
            self?.activities = data
            self?.presenter.presentLoadActivities(response: HomeBusinessModel.LoadData.ActivitiesResponse(items: data))
        }, failure: { [weak self] error in
            self?.presenter.removeLoading()
            self?.presenter.presentError(text: "Ha ocurrido un error al cargar los datos.")
        })
    }
    
    // Will get data from places according to the query
    func searchPlaces(query: String) {
        presenter.presentLoading()
        worker.searchPlaces(
            query: query,
            success: { [weak self] data in
                print(data)
                self?.places = data
                self?.presenter.removeLoading()
                self?.presenter.presentLoadPlaces(response: HomeBusinessModel.LoadData.PlacesResponse(items: data))
            }, failure: { [weak self] error in
                print(error!)
                self?.presenter.removeLoading()
                self?.presenter.presentError(text: "Ha ocurrido un error al cargar los datos.")
            }
        )
    }
    
    // Will get data from activities according to the query
    func searchActivities(query: String) {
        presenter.presentLoading()
        worker.searchActivities(
            query: query,
            success: { [weak self] data in
                print("Interactor activities")
                self?.activities = data
                self?.presenter.removeLoading()
                self?.presenter.presentLoadActivities(response: HomeBusinessModel.LoadData.ActivitiesResponse(items: data))
            }, failure: { [weak self] error in
                self?.presenter.removeLoading()
                self?.presenter.presentError(text: "Ha ocurrido un error al cargar los datos.")
            }
        )
    }
    
    // Will set the selected place to the datastore and present detail scene
    func selectPlace(request: HomeBusinessModel.SelectPlace.Request) {
        selectedPlace = places.first { $0.uid == request.uid }
        print("Selected place \(selectedPlace?.name)")
        presenter.presentRouteToDetail()
    }
    
    // will set the selected activity to the datastore and present detail scene
    func selectActivity(request: HomeBusinessModel.SelectActivity.Request) {
        selectedActivity = activities.first { $0.uid == request.uid }
        print("Selected act \(selectedActivity?.name)")
        presenter.presentRouteToDetail()
    }

    func deselect() {
        selectedPlace = nil
        selectedActivity = nil
    }
    
}
