//
//  HomePresenter.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import UIKit

//These Protocols are abstractions of presenter and are used by the interactor
//This protocol contains functions to load data and route to other scenes
protocol HomePresenterProtocol {
    func presentLoadPlaces(response: HomeBusinessModel.LoadData.PlacesResponse)
    func presentLoadActivities(response: HomeBusinessModel.LoadData.ActivitiesResponse)
    func presentRouteToDetail()
}
//This protocol has functions to present, hide activity indicator and present error dialog
protocol HomePresentAlertsProtocol{
    func presentLoading()
    func removeLoading()
    func presentError(text: String)
}

//The presenter will transform the information into the viewmodel and pass it to the viewcontroller
//Also will be in charge of caching images
class HomePresenter: HomePresenterProtocol, HomePresentAlertsProtocol {
    
    //Viewcontroller reference
    weak var viewController: (
        HomeViewControllerProtocol &
        HomeViewControllerAlertsProtocol &
        HomeViewControllerRouteProtocol
    )?
    
    //Initializer
    let imageCache: ImageCacheManager
    init(imageCache: ImageCacheManager){
        self.imageCache = imageCache
    }
    
    //This function will get the places data from interactor
    //Will convert it to the viewmodel
    //Will pass it to the viewController
    func presentLoadPlaces(response: HomeBusinessModel.LoadData.PlacesResponse) {
        viewController?.displayLoadData(
            viewModel: .init(
                items: response.items.map { placesModelToViewModel($0) }
            )
        )
    }
    
    //This function will get the activities data from interactor
    //Will convert it to the viewmodel
    //Will pass it to the viewController
    func presentLoadActivities(response: HomeBusinessModel.LoadData.ActivitiesResponse) {
        viewController?.displayLoadData(
            viewModel: .init(
                items: response.items.map { activitiesModelToViewModel($0) }
            )
        )
    }
    
    //Convert places models to item viewmodel
    func placesModelToViewModel(_ model: Place) -> HomeBusinessModel.LoadData.ItemViewModel {
        .init(
            title1: model.name,
            text1: "En tu ciudad",
            text2: model.tags,
            uid: model.uid,
            image: "",
            loadImage: { [weak self] imageView in
                self?.imageCache.getImage(model.image){ image in
                    imageView.image = image
                }
            }
        )
    }
    
    //Convert activities models to item viewmodel
    func activitiesModelToViewModel(_ model: Activity) -> HomeBusinessModel.LoadData.ItemViewModel {
        .init(
            title1: model.name,
            text1: model.dateTime,
            text2: model.tags,
            uid: model.uid,
            image: "",
            loadImage: { imageView in // todo
                imageView.image = UIImage(named: "QuillairRunner.png")
            }
        )
    }
    
    func presentLoading() {
        viewController?.displayLoading()
    }
    func removeLoading() {
        viewController?.hideLoading()
    }
    func presentRouteToDetail() {
        viewController?.displayRouteToDetail()
    }
    func presentError(text: String) {
        viewController?.displayError(text: text)
    }
    
    
}
