//
//  HomeViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import UIKit

protocol HomeViewControllerProtocol: AnyObject {
    func displayLoadData(viewModel: HomeBusinessModel.LoadData.ViewModel)
}

protocol HomeViewControllerRouteProtocol: AnyObject {
    func displayRouteToDetail()
}

protocol HomeViewControllerAlertsProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
    func displayError(text: String)
}

class HomeViewController: UIViewController,
                          UITableViewDataSource,
                          UITableViewDelegate,
                          UISearchBarDelegate,
                          HomeViewControllerProtocol,
                          HomeViewControllerRouteProtocol,
                          HomeViewControllerAlertsProtocol {
    
    @IBOutlet weak var getLocationBtn: UIButton!
    @IBOutlet weak var searchbar: UISearchBar!
    @IBOutlet weak var exploreSegment: UISegmentedControl!
    @IBOutlet weak var tableView: UITableView!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    
    var items: [HomeBusinessModel.LoadData.ItemViewModel] = []
    
    let interactor: HomeInteractorProtocol
    let router: HomeRouterProtocol
    
    var isPlaces = false
    
    init(interactor: HomeInteractorProtocol, router: HomeRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: HomeViewController.self)
        super.init(nibName: "\(HomeViewController.self)", bundle: bundle)
    }
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        self.tableView.dataSource = self
        self.tableView.delegate = self
        self.searchbar.delegate = self
        let nib = UINib(nibName: "HomeTableViewCell", bundle: nil)
        self.tableView.register(nib, forCellReuseIdentifier: "HomeTableViewCell")
        // register all type of cell you wanted to show.
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        navigationController?.setNavigationBarHidden(true, animated: animated)
        if(isPlaces){
            interactor.getPlaces()
        }else{
            interactor.getActivities()
        }
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        navigationController?.setNavigationBarHidden(false, animated: animated)
    }
    
    //button to navigate to add place
    @IBAction func locationClick(_ sender: Any) {
        router.routeToAddPlace()
    }
    
    //Table delegates
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return items.count
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "HomeTableViewCell", for: indexPath) as! HomeTableViewCell
        let item = items[indexPath.row]
        cell.setupData(viewModel: item)
        return cell
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return 95
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        interactor.deselect()
        let item = items[indexPath.row]
        if(isPlaces){
            interactor.selectPlace(request: .init(uid: item.uid))
        }else{
            interactor.selectActivity(request: .init(uid: item.uid))
        }
    }
    
    //Search bar functions
    //Will search automatically after user writes
    //When bar empty requests the original list again
    func searchBar(_ searchBar: UISearchBar, textDidChange searchText: String) {
        NSObject.cancelPreviousPerformRequests(
            withTarget: self,
            selector: #selector(self.reload(_:)),
            object: searchBar)
        perform(#selector(self.reload(_:)), with: searchBar, afterDelay: 0.75)
    }
    @objc func reload(_ searchBar: UISearchBar) {
        guard let query = searchBar.text, query.trimmingCharacters(in: .whitespaces) != "" else {
            print("nothing to search")
            if(isPlaces){
                interactor.getPlaces()
            }else{
                interactor.getActivities()
            }
            return
        }
        print(query)
        if(isPlaces){
            interactor.searchPlaces(query: query)
        }else{
            interactor.searchActivities(query: query)
        }
    }
    
    //used by the presenter to send data to the view
    func displayLoadData(viewModel: HomeBusinessModel.LoadData.ViewModel) {
        self.items = viewModel.items
        self.tableView.reloadData()
    }
    
    // tell router to show detail scene
    func displayRouteToDetail() {
        router.routeToDetail()
    }

    func displayLoading() {
        loader.isHidden = false
        tableView.isHidden = true
    }
    
    func hideLoading() {
        loader.isHidden = true
        tableView.isHidden = false
    }
    
    //Show error dialog
    func displayError(text: String) {
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Recargar", style: .default, handler: { (action: UIAlertAction!) in
            if(self.isPlaces){
                self.interactor.getPlaces()
            }else{
                self.interactor.getActivities()
            }
        }))
        self.present(alert, animated: true, completion: nil)
    }
    
    //Requestst data depending on the selected option from the segment
    @IBAction func selectOption(_ sender: Any) {
        switch exploreSegment.selectedSegmentIndex
        {
        case 0:
            //call interactor to show activities
            interactor.getActivities()
            isPlaces = false
        case 1:
            //call interactor to show places
            interactor.getPlaces()
            isPlaces = true
        default:
            break
        }
    }
    
    
}
