//
//  HomeWorker.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import Foundation

// Worker protocol is an abstraction of the class
protocol HomeWorkerProtocol {
    func getPlaces(
        success: @escaping (_ model: [Place]) -> Void,
        failure: @escaping (_ error: Error?) -> Void)
    func getActivities(
        success: @escaping (_ model: [Activity]) -> Void,
        failure: @escaping (_ error: Error?) -> Void)
    func searchPlaces(
        query: String,
        success: @escaping (_ model: [Place]) -> Void,
        failure: @escaping (_ error: Error?) -> Void)
    func searchActivities(
        query: String,
        success: @escaping (_ model: [Activity]) -> Void,
        failure: @escaping (_ error: Error?) -> Void)
}

// Worker: This class is in charge of the api calls for the scene
class HomeWorker: HomeWorkerProtocol {

    let firestore: FirebaseFirestoreManager
    let location: LocationManager
    
    //To search for a place or an activity we need the city
    //To get the user city we need to request location
    //Then we need to do reverse geolocation to know the name
    
    init(firestore: FirebaseFirestoreManager, location: LocationManager){
        self.firestore = firestore
        self.location = location
    }
    
    //Will get all the places in a city
    func getPlaces(
        success: @escaping (_ model: [Place]) -> Void,
        failure: @escaping (_ error: Error?) -> Void) {
        
        //get user current position
        location.fetchLocation(){ [weak self] position, error in
            
            if(error != nil){
                failure(error)
                return
            }
            //get city from reverse geolocation
            self?.location.getCityFromLocation(location: position!){ [weak self] city in
                
                var defaultCity = "barranquilla"
                if let city = city {
                    defaultCity = city
                }
                
                //get places from city
                self?.firestore.fetchDeeper(
                    mainCollection: "Cities",
                    document: defaultCity,
                    secondaryCollection: "Places"){ data, error in
                    
                    if error != nil {
                        failure(error)
                        return
                    }
                    var places: [Place] = []
                    for doc in data {
                        print("\(doc.documentID) => \(doc.data())")
                        let docData = doc.data()
                        places.append(
                            Place(
                                lat: docData["lat"] as! Double,
                                lng: docData["lng"] as! Double,
                                name: docData["name"] as! String,
                                tags: docData["tags"] as! String,
                                description: docData["description"] as! String,
                                image: docData["image"] as! String,
                                uid: doc.documentID
                            )
                        )
                    }
                    print("places: \(places.count)")
                    success(places)
                }
            }
        }
        

    }
    
    //Will get all activities in the city
    func getActivities(
        success: @escaping ([Activity]) -> Void,
        failure: @escaping (Error?) -> Void) {
        
        //get user current position
        location.fetchLocation(){ [weak self] position, error in
            
            if(error != nil){
                failure(error)
                return
            }
            //get city from reverse geolocation
            self?.location.getCityFromLocation(location: position!){ [weak self] city in
                var defaultCity = "barranquilla"
                if let city = city {
                    defaultCity = city
                }
                //get activities from city
                self?.firestore.fetchDeeper(
                    mainCollection: "Cities",
                    document: defaultCity,
                    secondaryCollection: "Activities"){ data, error in
                    
                    if error != nil {
                        failure(error)
                        return
                    }
                    var activities: [Activity] = []
                    for doc in data {
                        print("\(doc.documentID) => \(doc.data())")
                        let docData = doc.data()
                        activities.append(
                            Activity(
                                lat: docData["lat"] as! Double,
                                lng: docData["lng"] as! Double,
                                name: docData["name"] as! String,
                                description: docData["description"] as! String,
                                dateTime: docData["dateTime"] as! String,
                                placeName: docData["placeName"] as! String,
                                placeUid: docData["placeUid"] as! String,
                                tags: docData["tags"] as! String,
                                attendees: docData["attendees"] as! Int,
                                attendeesIDs: docData["attendeesIDs"] as! [String],
                                creator: docData["creator"] as! String,
                                isPublic: docData["isPublic"] as! Bool,
                                uid: doc.documentID
                            )
                        )
                    }
                    print("activities: \(activities.count)")
                    success(activities)
                }

            }
        }
    }

    func searchPlaces(query: String, success: @escaping ([Place]) -> Void, failure: @escaping (Error?) -> Void) {
        
        //get user current position
        location.fetchLocation(){ [weak self] position, error in
            
            if(error != nil){
                failure(error)
                return
            }
            //get city from reverse geolocation
            self?.location.getCityFromLocation(location: position!){ [weak self] city in
                
                var defaultCity = "barranquilla"
                if let city = city {
                    defaultCity = city
                }
                
                //get places from city
                self?.firestore.searchDeeper(
                    query: query,
                    field: "name",
                    mainCollection: "Cities",
                    document: defaultCity,
                    secondaryCollection: "Places"){ data, error in
                    
                    if error != nil {
                        failure(error)
                        return
                    }
                    var places: [Place] = []
                    for doc in data {
                        print("\(doc.documentID) => \(doc.data())")
                        let docData = doc.data()
                        places.append(
                            Place(
                                lat: docData["lat"] as! Double,
                                lng: docData["lng"] as! Double,
                                name: docData["name"] as! String,
                                tags: docData["tags"] as! String,
                                description: docData["description"] as! String,
                                image: docData["image"] as! String,
                                uid: doc.documentID
                            )
                        )
                    }
                    print("places: \(places.count)")
                    success(places)
                }
            }
        }
    }
    
    func searchActivities(query: String, success: @escaping ([Activity]) -> Void, failure: @escaping (Error?) -> Void) {
        
        //get user current position
        location.fetchLocation(){ [weak self] position, error in
            
            if(error != nil){
                failure(error)
                return
            }
            //get city from reverse geolocation
            self?.location.getCityFromLocation(location: position!){ [weak self] city in
                var defaultCity = "barranquilla"
                if let city = city {
                    defaultCity = city
                }
                //get activities from city
                self?.firestore.searchDeeper(
                    query: query,
                    field: "name",
                    mainCollection: "Cities",
                    document: defaultCity,
                    secondaryCollection: "Activities"){ data, error in
                    
                    if error != nil {
                        failure(error)
                        return
                    }
                    var activities: [Activity] = []
                    for doc in data {
                        print("\(doc.documentID) => \(doc.data())")
                        let docData = doc.data()
                        activities.append(
                            Activity(
                                lat: docData["lat"] as! Double,
                                lng: docData["lng"] as! Double,
                                name: docData["name"] as! String,
                                description: docData["description"] as! String,
                                dateTime: docData["dateTime"] as! String,
                                placeName: docData["placeName"] as! String,
                                placeUid: docData["placeUid"] as! String,
                                tags: docData["tags"] as! String,
                                attendees: docData["attendees"] as! Int,
                                attendeesIDs: docData["attendeesIDs"] as! [String],
                                creator: docData["creator"] as! String,
                                isPublic: docData["isPublic"] as! Bool,
                                uid: doc.documentID
                            )
                        )
                    }
                    print("activities: \(activities.count)")
                    success(activities)
                }

            }
        }
    }
    
    func getLocation() {
        location.fetchLocation(){ location, error in
            print("location: \(String(describing: location))")
        }
    }
    
}
