//
//  LoginViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import UIKit

//Protocol to route back to root
protocol LoginViewControllerRouteProtocol: AnyObject {
    func displayRouteToRoot()
    func displaySetupView(viewModel: LoginBusinessModel.SetupView.ViewModel)
    func displayError(text: String)
}

//Protocol to display loading
protocol LoginViewControllerLoadingProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
}

//View controller for login
class LoginViewController: UIViewController,
                           LoginViewControllerRouteProtocol,
                           LoginViewControllerLoadingProtocol {

    //View components
    @IBOutlet weak var userInput: UITextField!
    @IBOutlet weak var passwordInput: UITextField!
    @IBOutlet weak var validationText: UILabel!
    @IBOutlet weak var sessionIssuesBtn: UIButton!
    @IBOutlet weak var registerBtn: UIButton!
    @IBOutlet weak var signInBtn: UIButton!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    
    var viewModel: LoginBusinessModel.SetupView.ViewModel?
    
    let interactor: LoginInteractorProtocol
    let router: LoginRouterProtocol
    
    init(interactor: LoginInteractorProtocol, router: LoginRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: LoginViewController.self)
        super.init(nibName: "\(LoginViewController.self)", bundle: bundle)
    }
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        interactor.viewDidLoad()
    }
    
    @IBAction func loginProblemsClick(_ sender: Any) {
        router.routeToPassReset()
    }
    @IBAction func registerClick(_ sender: Any) {
        router.routeToRegister()
    }
    @IBAction func signInClick(_ sender: Any) {
        
        guard let validator = viewModel?.validator(userInput.text, passwordInput.text) else{
            return
        }
        validationText.isHidden = validator.isValid
        validationText.text = validator.errorMessage
        
        if validator.isValid {
            interactor.login(
                request: LoginBusinessModel.Login.Request(
                    username: validator.usernameValue,
                    password: validator.passwordValue
                )
            )
        }
    }
    
    // Protocol stubs
    // tell router to go to rootViewController
    func displayRouteToRoot() {
        router.routeToRoot()
    }
    
    // add stylization to view
    func displaySetupView(viewModel: LoginBusinessModel.SetupView.ViewModel) {
        self.viewModel = viewModel
        signInBtn.layer.cornerRadius = viewModel.signIn.cornerRadius
        signInBtn.layer.borderWidth = viewModel.signIn.borderWidth
        signInBtn.layer.borderColor = viewModel.signIn.borderColor
        self.navigationController?.setNavigationBarHidden(true, animated: false)
        //Looks for single or multiple taps.
        let tap = UITapGestureRecognizer(
            target: self,
            action: #selector(UIInputViewController.dismissKeyboard)
        )
        view.addGestureRecognizer(tap)
    }
    
    //Calls this function when the tap is recognized.
    @objc func dismissKeyboard() {
        //Causes the view (or one of its embedded text fields) to resign the first responder status.
        view.endEditing(true)
    }
    
    // display activity indicator
    func displayLoading() {
        signInBtn.isHidden = true
        registerBtn.isHidden = true
        sessionIssuesBtn.isHidden = true
        validationText.isHidden = true
        loader.isHidden = false
    }
    
    // remove activity indicator
    func hideLoading() {
        signInBtn.isHidden = false
        registerBtn.isHidden = false
        sessionIssuesBtn.isHidden = false
        loader.isHidden = true
    }
    
    func displayError(text: String){
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Entendido", style: .default, handler: nil))
        self.present(alert, animated: true, completion: nil)
    }
    
}
