//
//  MyEventsViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez on 1/12/21.
//

import UIKit

protocol MyEventsViewControllerProtocol: AnyObject {
    func displayLoadData(viewModel: MyEventsBusinessModel.LoadData.ViewModel)
}

protocol MyEventsViewControllerAlertsProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
    func displayError(text: String)
}


class MyEventsViewController: UIViewController,
                              UITableViewDataSource,
                              UITableViewDelegate,
                              UISearchBarDelegate,
                              MyEventsViewControllerProtocol,
                              MyEventsViewControllerAlertsProtocol {
    
    
    @IBOutlet weak var searchBar: UISearchBar!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    @IBOutlet weak var table: UITableView!
    
    var items: [MyEventsBusinessModel.LoadData.ItemViewModel] = []
    
    let interactor: MyEventsInteractorProcotol
    let router: MyEventsRouterProtocol
    
    init(interactor: MyEventsInteractorProcotol, router: MyEventsRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: MyEventsViewController.self)
        super.init(nibName: "\(MyEventsViewController.self)", bundle: bundle)
    }
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()
        //Tell interactor to get events data
        interactor.getEvents()
        self.table.dataSource = self
        self.table.delegate = self
        self.searchBar.delegate = self
        let nib = UINib(nibName: "MyEventsTableViewCell", bundle: nil)
        self.table.register(nib, forCellReuseIdentifier: "MyEventsTableViewCell")
    }
    
    //Remove navigation bar
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        navigationController?.setNavigationBarHidden(true, animated: animated)
    }
    //Show navigation bar
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        navigationController?.setNavigationBarHidden(false, animated: animated)
    }
    
    func displayLoadData(viewModel: MyEventsBusinessModel.LoadData.ViewModel) {
        self.items = viewModel.items
        self.table.reloadData()
    }
    
    func displayLoading() {
        table.isHidden = true
        loader.isHidden = false
    }
    
    func hideLoading() {
        loader.isHidden = true
        table.isHidden = false
    }
    
    func displayError(text: String) {
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Recargar", style: .default, handler: { (action: UIAlertAction!) in
            //Tell interactor to reload
            self.interactor.getEvents()
        }))
        self.present(alert, animated: true, completion: nil)
    }
    
    //Table delegates
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return items.count
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "MyEventsTableViewCell", for: indexPath) as! MyEventsTableViewCell
        let item = items[indexPath.row]
        cell.setupData(viewModel: item)
        return cell
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return 95
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        let item = items[indexPath.row]
        //Show alert that will ask to update the event (say that will not assist anymore)
        let changeEventAlert = UIAlertController(
            title: "Cambiar Asistencia",
            message: "Si ya no quieres asistir a este evento haz click en el botón correspondiente. Si haces esto el evento se eliminará de tu lista.",
            preferredStyle: UIAlertController.Style.alert)

        changeEventAlert.addAction(UIAlertAction(title: "No Asistiré", style: .default, handler: { (action: UIAlertAction!) in
            //if clicks on not assist tell interactor to updateEvent
            self.interactor.updateEventsActivities(
                request: .init(uid: item.uid, city: item.city, isCreator: item.isCreator)
            )
        }))
        changeEventAlert.addAction(UIAlertAction(title: "Cancelar", style: .cancel, handler: nil))
        present(changeEventAlert, animated: true, completion: nil)
    }
    
    //Search bar functions
    //Will search automatically after user writes
    //When bar empty requests the original list again
    func searchBar(_ searchBar: UISearchBar, textDidChange searchText: String) {
        NSObject.cancelPreviousPerformRequests(
            withTarget: self,
            selector: #selector(self.reload(_:)),
            object: searchBar)
        perform(#selector(self.reload(_:)), with: searchBar, afterDelay: 0.75)
    }
    @objc func reload(_ searchBar: UISearchBar) {
        guard let query = searchBar.text, query.trimmingCharacters(in: .whitespaces) != "" else {
            print("nothing to search")
            interactor.getEvents()
            return
        }
        print(query)
        interactor.searchEvents(query: query)
    }
    
    
    
}
