//
//  MyEventsWorker.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez on 1/12/21.
//

import Foundation

//Worker abtraction
//This class will be used by the interactor
protocol MyEventsWorkerProtocol {
    func getEvents(
        success: @escaping (_ model: [UserEvent]) -> Void,
        failure: @escaping (_ error: Error?) -> Void)
    func searchEvents(
        query: String,
        success: @escaping (_ model: [UserEvent]) -> Void,
        failure: @escaping (_ error: Error?) -> Void)
    func updateAttendeesCount(
        uid: String,
        city: String,
        isCreator: Bool,
        completion: @escaping (_ didUpdate: Bool)->Void)
}

//Will contain requests to apis and services
class MyEventsWorker: MyEventsWorkerProtocol {

    let firestore: FirebaseFirestoreManager
    let location: LocationManager
    let auth: FirebaseAuthManager
    
    init(firestore: FirebaseFirestoreManager, location: LocationManager, auth: FirebaseAuthManager){
        self.firestore = firestore
        self.location = location
        self.auth = auth
    }
    
    func getEvents(
        success: @escaping ([UserEvent]) -> Void,
        failure: @escaping (Error?) -> Void) {
        
        //Get user id
        let userUID = auth.getUser()!.uid as String
        //get events from user
        firestore.fetchDeeper(
            mainCollection: "Users",
            document: userUID,
            secondaryCollection: "Events"){ data, error in
            
            if error != nil {
                failure(error)
                return
            }
            var events: [UserEvent] = []
            for doc in data {
                print("\(doc.documentID) => \(doc.data())")
                let docData = doc.data()
                events.append(
                    UserEvent(
                        activityUid: docData["activityUid"] as! String,
                        date: docData["date"] as! String,
                        isCreator: docData["isCreator"] as! Bool,
                        isGoing: docData["isGoing"] as! Bool,
                        eventName: docData["name"] as! String,
                        placeName: docData["where"] as! String,
                        city: docData["city"] as! String)
                )
            }
            print("events: \(events.count)")
            success(events)
        }
        
    }
    
    func searchEvents(
        query: String,
        success: @escaping ([UserEvent]) -> Void,
        failure: @escaping (Error?) -> Void) {
        
        //Get user id
        let userUID = auth.getUser()!.uid as String
        
        firestore.searchDeeper(
            query: query,
            field: "name",
            mainCollection: "Users",
            document: userUID,
            secondaryCollection: "Events"){ data, error in
            
            if error != nil {
                failure(error)
                return
            }
            var events: [UserEvent] = []
            for doc in data {
                print("\(doc.documentID) => \(doc.data())")
                let docData = doc.data()
                events.append(
                    UserEvent(
                        activityUid: docData["activityUid"] as! String,
                        date: docData["date"] as! String,
                        isCreator: docData["isCreator"] as! Bool,
                        isGoing: docData["isGoing"] as! Bool,
                        eventName: docData["name"] as! String,
                        placeName: docData["where"] as! String,
                        city: docData["city"] as! String)
                )
            }
            print("events: \(events.count)")
            success(events)
            
        }
    }
    
    //If is creator will delete the event from my events and the activity from the home
    //If isnt creator will edit the activity and delete the event from my events
    func updateAttendeesCount(
        uid: String, //Activity uid
        city: String, //City for looking into the collection
        isCreator: Bool,
        completion: @escaping (_ didUpdate: Bool)->Void){
        
        let userUID = auth.getUser()!.uid as String
        
        if(isCreator){
            //If is creator remove the activity
            //remove the event from the list
            firestore.deleteFromDeeper(
                mainCollection: "Users",
                mainDocument: userUID,
                secondaryCollection: "Events",
                document: uid){ [weak self] didDelete in
                if(didDelete){
                    //remove the event from the list
                    self?.firestore.deleteFromDeeper(
                        mainCollection: "Cities",
                        mainDocument: city,
                        secondaryCollection: "Activities",
                        document: uid){ didDelete in
                        completion(didDelete)
                    }
                }else{
                    completion(false)
                }
            }

        }else{
            //First get the activity data
            firestore.fetchDocumentDeeper(
                mainCollection: "Cities",
                mainDocument: city,
                secondaryCollection: "Activities",
                document: uid){ [weak self] document, error in
                
                if error != nil {
                    completion(false)
                }else{
                    //get the count and the array
                    let docData = document?.data()
                    let attendees = docData!["attendees"] as! Int
                    let attendeesIDs = docData!["attendeesIDs"] as! [String]
                    //modify the count and the array
                    let updatedAttendees = attendees - 1
                    let updatedAttendeesIDs = attendeesIDs.filter { $0 != userUID }
                    let d:[String : Any] = [
                        "attendees": updatedAttendees,
                        "attendeesIDs": updatedAttendeesIDs
                    ]
                    //update the activity with the new count and the new array
                    self?.firestore.updateDeeper(
                        mainCollection: "Cities",
                        mainDocument: city,
                        secondaryCollection: "Activities",
                        secondaryDocument: uid,
                        data: d){ [weak self] didUpdate in
                        if(didUpdate){
                            //remove the event from the list
                            self?.firestore.deleteFromDeeper(
                                mainCollection: "Users",
                                mainDocument: userUID,
                                secondaryCollection: "Events",
                                document: uid){ didDelete in
                                completion(didDelete)
                            }
                        }else{
                            completion(false)
                        }
                        
                    }
        
                }
                
            }
        }
        
      
 
        
    }
    
    
    
}
