//
//  MyProfileViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez on 1/12/21.
//

import UIKit
import MobileCoreServices

//Protocol used by the presenter to pass data
protocol MyProfileViewControllerProtocol: AnyObject {
    func displayData(viewModel: MyProfileBusinessModel.FetchUser.ViewModel)
    func displaySetupView(viewModel: MyProfileBusinessModel.SetupView.ViewModel)
    func displayResultAlert(text: String)
    func displayWarningAlert(text: String)
    func displayRouteToRoot()
}
//Protocol to display loading
protocol MyProfileViewLoadingProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
}

class MyProfileViewController: UIViewController,
                               UINavigationControllerDelegate,
                               UIImagePickerControllerDelegate,
                               MyProfileViewControllerProtocol,
                               MyProfileViewLoadingProtocol {

    @IBOutlet weak var photoView: UIImageView!
    @IBOutlet weak var tagsText: UILabel!
    @IBOutlet weak var closeSession: UIButton!
    @IBOutlet weak var changeImageBtn: UIButton!
    @IBOutlet weak var editPreferencesBtn: UIButton!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    @IBOutlet weak var emailText: UILabel!
    
    var viewModel: MyProfileBusinessModel.SetupView.ViewModel?
    var imageURL: URL? = nil
    var interests: String = ""
    
    let interactor: MyProfileInteractorProtocol
    let router: MyProfileRouterProtocol
    
    init(interactor: MyProfileInteractorProtocol, router: MyProfileRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: MyProfileViewController.self)
        super.init(nibName: "\(MyProfileViewController.self)", bundle: bundle)
    }
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        interactor.viewDidLoad()
    }
    
    func displayData(viewModel: MyProfileBusinessModel.FetchUser.ViewModel) {
        tagsText.text = viewModel.tags
        emailText.text = viewModel.email
        interests = viewModel.tags
        viewModel.loadImage(photoView)
    }
    
    func displaySetupView(viewModel: MyProfileBusinessModel.SetupView.ViewModel) {
        self.viewModel = viewModel
        closeSession.layer.cornerRadius = viewModel.grayButton.cornerRadius
        closeSession.layer.borderWidth = viewModel.grayButton.borderWidth
        closeSession.layer.borderColor = viewModel.grayButton.borderColor
        editPreferencesBtn.layer.cornerRadius = viewModel.greenButton.cornerRadius
        editPreferencesBtn.layer.borderWidth = viewModel.greenButton.borderWidth
        editPreferencesBtn.layer.borderColor = viewModel.greenButton.borderColor
        changeImageBtn.layer.cornerRadius = viewModel.greenButton.cornerRadius
        changeImageBtn.layer.borderWidth = viewModel.greenButton.borderWidth
        changeImageBtn.layer.borderColor = viewModel.greenButton.borderColor
    }
    
    func displayResultAlert(text: String) {
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Entendido", style: .default, handler: { (action: UIAlertAction!) in
            self.interactor.getUserData()
        }))
        self.present(alert, animated: true, completion: nil)
    }
    
    //Show alert before closing session
    func displayWarningAlert(text: String) {
        let warningAlert = UIAlertController(
            title: "Aviso",
            message: "¿Esta seguro que quiere cerrar la sesión?",
            preferredStyle: UIAlertController.Style.alert)

        warningAlert.addAction(UIAlertAction(title: "Cerrar sesión", style: .default, handler: { (action: UIAlertAction!) in
            //if clicks on not assist tell interactor to close session
            self.interactor.logout()
        }))
        warningAlert.addAction(UIAlertAction(title: "Cancelar", style: .cancel, handler: nil))
        present(warningAlert, animated: true, completion: nil)
    }
    
    func displayLoading() {
        loader.isHidden = false
        changeImageBtn.isHidden = true
        editPreferencesBtn.isHidden = true
        closeSession.isHidden = true
    }
    
    func hideLoading() {
        loader.isHidden = true
        changeImageBtn.isHidden = false
        editPreferencesBtn.isHidden = false
        closeSession.isHidden = false
    }
    
    //Remove navigation bar
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        navigationController?.setNavigationBarHidden(true, animated: animated)
        interactor.getUserData()
    }
    //Show navigation bar
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        navigationController?.setNavigationBarHidden(false, animated: animated)
    }
    
    @IBAction func changePhotoAction(_ sender: Any) {
        let vc = UIImagePickerController()
        vc.allowsEditing = true
        vc.delegate = self
        if UIImagePickerController.isSourceTypeAvailable(.camera) {
            vc.sourceType = .camera
        }else{
            vc.sourceType = .photoLibrary
        }
        present(vc, animated: true)
    }
    func imagePickerController(_ picker: UIImagePickerController, didFinishPickingMediaWithInfo info: [UIImagePickerController.InfoKey : Any]) {
        picker.dismiss(animated: true)

        guard let image = info[.editedImage] as? UIImage else {
            print("No image found")
            return
        }
        photoView.image = image
        
        // Check for the media type
        let mediaType = info[UIImagePickerController.InfoKey.mediaType] as! CFString
        if mediaType == kUTTypeImage {
            imageURL = info[UIImagePickerController.InfoKey.imageURL] as? URL
            interactor.updateUserImage(request: .init(path: imageURL!))
        }
    }
    
    @IBAction func editTagsAction(_ sender: Any) {
        let alert = UIAlertController(
            title: "Editar intereses",
            message: "Escribe en el cuadro de texto tus nuevos intereses.",
            preferredStyle: .alert)
        alert.addTextField { (textField) in
            textField.placeholder = "Escribe aquí (patinar, nadar)"
            textField.text = self.interests
        }
        
        alert.addAction(UIAlertAction(title: "Guardar", style: .default, handler: { [weak alert] (_) in
            guard let textField = alert?.textFields?[0], let userText = textField.text else { return }
            self.interactor.updateUserTags(request: .init(tags: userText))
        }))
        alert.addAction(UIAlertAction(title: "Cancelar", style: .cancel, handler: nil))
        self.present(alert, animated: true, completion: nil)
    }
    
    func displayRouteToRoot() {
        router.routeToRoot()
    }
    
    @IBAction func logoutAction(_ sender: Any) {
        interactor.logout()
    }
    
    
}

