//
//  NewEventViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import UIKit
import MapKit

//Protocol used by the presenter to pass data
protocol NewEventViewControllerProtocol: AnyObject {
    func displayData(viewModel: NewEventBusinessModel.LoadData.ViewModel)
    func displaySetupView(viewModel: NewEventBusinessModel.SetupView.ViewModel)
    func displayResultAlert(text: String)
}
//Protocol to display loading
protocol NewEventViewLoadingProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
}

class NewEventViewController: UIViewController,
                              NewEventViewControllerProtocol,
                              NewEventViewLoadingProtocol {
    
    @IBOutlet weak var map: MKMapView!
    @IBOutlet weak var placeNameText: UILabel!
    @IBOutlet weak var nameEventInput: UITextField!
    @IBOutlet weak var descriptionInput: UITextField!
    @IBOutlet weak var tagsInput: UITextField!
    @IBOutlet weak var dateTimeInput: UIDatePicker!
    @IBOutlet weak var validationText: UILabel!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    @IBOutlet weak var saveBtn: UIButton!
    
    var viewModel: NewEventBusinessModel.SetupView.ViewModel?
    
    let interactor: NewEventInteractorProtocol
    let router: NewEventRouterProtocol
    
    init(interactor: NewEventInteractorProtocol, router: NewEventRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: NewEventViewController.self)
        super.init(nibName: "\(NewEventViewController.self)", bundle: bundle)
    }
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        self.title = "Nuevo Evento"
        super.viewDidLoad()
        interactor.viewDidLoad()
        interactor.loadData()
    }
    
    @IBAction func saveAction(_ sender: Any) {
        guard let validator = viewModel?.validator(
                nameEventInput.text,
                descriptionInput.text,
                tagsInput.text) else{
            return
        }
        validationText.isHidden = validator.isValid
        validationText.text = validator.errorMessage
        
        let dateStr = formatDate(date: dateTimeInput.date)
        
        if validator.isValid {
            interactor.saveActivity(
                request: .init(
                    eventName: validator.eventNameValue,
                    description: validator.descriptionValue,
                    dateTime: dateStr,
                    tags: validator.tagsValue
                )
            )
        }
    }
    
    func formatDate(date: Date)->String{
       let formatter = DateFormatter()
       formatter.dateFormat = "dd/MM/yyyy HH:mm"
       let strdate = formatter.string(from: date)
       return strdate
    }
    
    //Set map information
    //Set data from viewmodel to view
    func displayData(viewModel: NewEventBusinessModel.LoadData.ViewModel) {
        //Center map to location and add a pin
        let cloc = CLLocationCoordinate2D(
            latitude: viewModel.position.latitude,
            longitude: viewModel.position.longitude)
        let coordinateRegion = MKCoordinateRegion.init(center: cloc,
                                                       latitudinalMeters: 500,
                                                       longitudinalMeters: 500)
        map.setRegion(coordinateRegion, animated: true)
        let annotation = MKPointAnnotation()
        annotation.coordinate = cloc
        map.addAnnotation(annotation)
        placeNameText.text = "Llegar a \(viewModel.placeName)"
    }
    
    func displaySetupView(viewModel: NewEventBusinessModel.SetupView.ViewModel) {
        self.viewModel = viewModel
        saveBtn.layer.cornerRadius = viewModel.button.cornerRadius
        saveBtn.layer.borderWidth = viewModel.button.borderWidth
        saveBtn.layer.borderColor = viewModel.button.borderColor
        //Looks for single or multiple taps.
        let tap = UITapGestureRecognizer(
            target: self,
            action: #selector(UIInputViewController.dismissKeyboard)
        )
        view.addGestureRecognizer(tap)
    }
    
    //Calls this function when the tap is recognized.
    @objc func dismissKeyboard() {
        //Causes the view (or one of its embedded text fields) to resign the first responder status.
        view.endEditing(true)
    }
    
    func displayResultAlert(text: String) {
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Entendido", style: .default, handler: { [weak self] action in
            //Tell router to go to the root
            self?.router.routeToRoot()
        }))
        self.present(alert, animated: true, completion: nil)
    }
    
    func displayLoading() {
        loader.isHidden = false
        validationText.isHidden = true
        saveBtn.isHidden = true
    }
    
    func hideLoading() {
        loader.isHidden = true
        saveBtn.isHidden = false
    }
    
    
    
}
