//
//  RegisterViewController.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import UIKit

//Protocol to route back to root
protocol RegisterViewControllerRouteProtocol: AnyObject {
    func displayRouteToPermissions()
    func displaySetupView(viewModel: RegisterBusinessModel.SetupView.ViewModel)
    func displayError(text: String)
}

//Protocol to display loading
protocol RegisterViewControllerLoadingProtocol: AnyObject {
    func displayLoading()
    func hideLoading()
}

class RegisterViewController: UIViewController, RegisterViewControllerRouteProtocol, RegisterViewControllerLoadingProtocol {

    
    @IBOutlet weak var userInput: UITextField!
    @IBOutlet weak var passwordInput: UITextField!
    @IBOutlet weak var confirmPasswordInput: UITextField!
    @IBOutlet weak var validationTex: UILabel!
    @IBOutlet weak var goLoginBtn: UIButton!
    @IBOutlet weak var registerBtn: UIButton!
    @IBOutlet weak var loader: UIActivityIndicatorView!
    
    var viewModel: RegisterBusinessModel.SetupView.ViewModel?
    
    let interactor: RegisterInteractorProtocol
    let router: RegisterRouterProtocol
    
    init(interactor: RegisterInteractorProtocol, router: RegisterRouterProtocol) {
        self.interactor = interactor
        self.router = router
        let bundle = Bundle(for: RegisterViewController.self)
        super.init(nibName: "\(RegisterViewController.self)", bundle: bundle)
    }
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        interactor.viewDidLoad()
    }
    
    @IBAction func goLoginClick(_ sender: Any) {
        router.routeToLogin()
    }
    @IBAction func registerClick(_ sender: Any) {
        guard let validator = viewModel?.validator(userInput.text, passwordInput.text, confirmPasswordInput.text) else {
            return
        }
        validationTex.isHidden = validator.isValid
        validationTex.text = validator.errorMessage
        if validator.isValid {
            interactor.register(
                request: RegisterBusinessModel.Register.Request(
                    username: validator.usernameValue,
                    password: validator.passwordValue
                )
            )
        }
    }
    
    func displayRouteToPermissions() {
        router.routeToPermissions()
    }
    
    func displaySetupView(viewModel: RegisterBusinessModel.SetupView.ViewModel) {
        self.viewModel = viewModel
        registerBtn.layer.cornerRadius = viewModel.register.cornerRadius
        registerBtn.layer.borderWidth = viewModel.register.borderWidth
        registerBtn.layer.borderColor = viewModel.register.borderColor
        self.navigationController?.setNavigationBarHidden(true, animated: false)
        //Looks for single or multiple taps.
        let tap = UITapGestureRecognizer(
            target: self,
            action: #selector(UIInputViewController.dismissKeyboard)
        )
        view.addGestureRecognizer(tap)
    }
    
    //Calls this function when the tap is recognized.
    @objc func dismissKeyboard() {
        //Causes the view (or one of its embedded text fields) to resign the first responder status.
        view.endEditing(true)
    }
    
    func displayError(text: String) {
        let alert = UIAlertController(title: "Aviso", message: text, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "Entendido", style: .default, handler: nil))
        self.present(alert, animated: true, completion: nil)
    }
    
    func displayLoading() {
        loader.isHidden = false
        registerBtn.isHidden = true
        validationTex.isHidden = true
        goLoginBtn.isHidden = true
    }
    
    func hideLoading() {
        loader.isHidden = true
        registerBtn.isHidden = false
        goLoginBtn.isHidden = false
    }
    
    
}

