//
//  CoreLocationServices.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import Foundation
import CoreLocation

//A class that encapsulates location services
//Helps decouple it from viewController
class CoreLocationServices: NSObject, LocationManager, CLLocationManagerDelegate {
    
    private var didFetchLocation: FetchLocationCompletion?
    private var requestLocationAuthorizationCallback: ((CLAuthorizationStatus) -> Void)?
    
    private lazy var locationManager: CLLocationManager = {
        
        // Initialize Location Manager
        let locationManager = CLLocationManager()
        // Configure Location Manager
        locationManager.delegate = self
        return locationManager
        
    }()
    
    func fetchLocation(completion: @escaping FetchLocationCompletion) {
        if CLLocationManager.authorizationStatus() == .authorizedWhenInUse {
            // Store Reference to Completion
            self.didFetchLocation = completion
            // Fetch Location
            locationManager.requestLocation()
        } else {
            locationManager.requestWhenInUseAuthorization()
        }
    }
    
    func requestLocationPermission() {
        self.locationManager.delegate = self
        let currentStatus = CLLocationManager.authorizationStatus()

        // Only ask authorization if it was never asked before
        guard currentStatus == .notDetermined else { return }

        // Starting on iOS 13.4.0, to get .authorizedAlways permission, you need to
        // first ask for WhenInUse permission, then ask for Always permission to
        // get to a second system alert
        if #available(iOS 13.4, *) {
            self.requestLocationAuthorizationCallback = { status in
               if status == .authorizedWhenInUse {
                   self.locationManager.requestAlwaysAuthorization()
               }
           }
           self.locationManager.requestWhenInUseAuthorization()
        } else {
           self.locationManager.requestAlwaysAuthorization()
        }
    }
    
    func locationManager(_ manager: CLLocationManager, didChangeAuthorization status: CLAuthorizationStatus) {
        if status == .notDetermined {
            // Request Authorization
            locationManager.requestWhenInUseAuthorization()
        } else if status == .authorizedWhenInUse {
            // Fetch Location
            locationManager.requestLocation()
        } else {
            // Invoke Completion Handler
            didFetchLocation?(nil, .notAuthorizedToRequestLocation)
            // Reset Completion Handler
            didFetchLocation = nil
        }
    }

    func locationManager(_ manager: CLLocationManager, didUpdateLocations locations: [CLLocation]) {
        guard let location = locations.first else {
            return
        }
        // Invoke Completion Handler
        didFetchLocation?(Location(location: location), nil)
        // Reset Completion Handler
        didFetchLocation = nil
    }

    func locationManager(_ manager: CLLocationManager, didFailWithError error: Error) {
        print("Unable to Fetch Location (\(error))")
    }
    
    func getCityFromLocation(location: Location, completion: @escaping (_ cityName: String?)->Void) {
        let ceo: CLGeocoder = CLGeocoder()
        let loc: CLLocation = CLLocation(latitude: location.latitude, longitude: location.longitude)
        
        ceo.reverseGeocodeLocation(loc, completionHandler:
            {(placemarks, error) in
                     if (error != nil)
                     {
                        print("reverse geodcode fail: \(error!.localizedDescription)")
                     }
                     let pm = placemarks! as [CLPlacemark]
                     if pm.count > 0 {
                        let pm = placemarks![0]
                        print(pm.locality)
                        completion(pm.locality)
                   }
             }
        )
    }
    
    
}

fileprivate extension Location {

    init(location: CLLocation) {
        latitude = location.coordinate.latitude
        longitude = location.coordinate.longitude
    }

}
