//
//  FirebaseFirestoreServices.swift
//  Quillair
//
//  Created by Valery Patrizia Madiedo Gomez
//

import Foundation
import Firebase

class FirebaseFirestoreServices: FirebaseFirestoreManager {
    
    private let db = Firestore.firestore()
    
    func deleteFromDeeper(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        document: String,
        completion: @escaping (Bool)->Void)
     {
        db.collection(mainCollection)
            .document(mainDocument)
            .collection(secondaryCollection)
            .document(document).delete() { err in
            if let err = err {
                print("Error removing document: \(err)")
                completion(false)
            } else {
                print("Document successfully removed!")
                completion(true)
            }
        }
    }
    
    
    //For updating second level collections
    //such as events of users, or activies / places to cities
    func updateDeeper(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        secondaryDocument: String,
        data: [String: Any],
        completion: @escaping (Bool) -> Void){
        
        let ref = db.collection(mainCollection)
            .document(mainDocument)
            .collection(secondaryCollection)
            .document(secondaryDocument)
        
        ref.updateData(data) { err in
            if let err = err {
                print("Error updating document: \(err)")
                completion(false)
            } else {
                print("Document successfully updated")
                completion(true)
            }
        }
    }
    
    //To update a document in a fisrt level collection
    //like the user image
    func updateWithId(collection: String, document: String, data: [String: Any], completion: @escaping (Bool) -> Void) {
        let ref = db.collection(collection).document(document)
        ref.updateData(data) { err in
            if let err = err {
                print("Error updating document: \(err)")
                completion(false)
            } else {
                print("Document successfully updated")
                completion(true)
            }
        }
    }
    
    // Will post data to a second level collection
    // Needs primary collection, document name and secondary collection name
    // Generates an id for the document
    // Used to post events for users, activities or places
    func postDataDeeper(
        mainCollection: String,
        document: String,
        secondaryCollection: String,
        data: [String : Any],
        completion: @escaping (_ isSuccessful: Bool, _ ref: String) -> Void) {
        // Add a new document with a generated ID
        var ref: DocumentReference? = nil
        ref = db.collection(mainCollection)
            .document(document)
            .collection(secondaryCollection)
            .addDocument(data: data) { err in
            if let err = err {
                print("Error adding document: \(err)")
                completion(false, "")
            } else {
                print("Document added with ID: \(ref!.documentID)")
                completion(true, ref!.documentID)
            }
        }
    }
    
    //For posting data to a nested collection (second level) with a document id
    //Used to save data of user events, places or activities
    func postDataDeeperWithUID(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        documentUID: String,
        data: [String : Any],
        completion: @escaping (_ isSuccessful: Bool) -> Void) {
        
        // Add a new document with a designated id
        db.collection(mainCollection)
            .document(mainDocument)
            .collection(secondaryCollection)
            .document(documentUID)
            .setData(data) { err in
            if err != nil {
                completion(false)
            } else {
                completion(true)
            }
        }
    }
    
    
    
    //Will post data to a first level collection
    //Needs an specific document name
    //Is used for posting users into the db
    func postDataWithDoc(collection: String, document: String, data: [String : Any], completion: @escaping (Bool) -> Void) {
        db.collection(collection).document(document).setData(data) { err in
            if let err = err {
                print("Error adding document: \(err)")
                completion(false)
            } else {
                print("Document added")
                completion(true)
            }
        }
    }
    
    //For getting documents in a second level collection
    //Which is why we'll use it for retrieving all activities or places
    func fetchDeeper(
        mainCollection: String,
        document: String,
        secondaryCollection: String,
        completion: @escaping ([QueryDocumentSnapshot], Error?) -> Void) {
        
        db.collection(mainCollection)
            .document(document)
            .collection(secondaryCollection)
            .getDocuments() { (querySnapshot, err) in
            if let err = err {
                print("Error getting documents: \(err)")
                DispatchQueue.main.async {
                    completion(querySnapshot!.documents, err)
                }
            } else {
                for document in querySnapshot!.documents {
                    print("\(document.documentID) => \(document.data())")
                }
                DispatchQueue.main.async {
                    completion(querySnapshot!.documents, err)
                }
            }
        }
        
    }
    
    //For getting documents that match the query
    //This will search in a second level collection
    //Which is why we'll use it for activities or places search
    func searchDeeper(
        query: String,
        field: String,
        mainCollection: String,
        document: String,
        secondaryCollection: String,
        completion: @escaping ([QueryDocumentSnapshot], Error?) -> Void) {
        db.collection(mainCollection)
            .document(document)
            .collection(secondaryCollection)
            .whereField(field, isGreaterThanOrEqualTo: query)
            .getDocuments() { (querySnapshot, err) in
            if let err = err {
                print("Error getting documents: \(err)")
                DispatchQueue.main.async {
                    completion(querySnapshot!.documents, err)
                }
            } else {
                for document in querySnapshot!.documents {
                    print("\(document.documentID) => \(document.data())")
                }
                DispatchQueue.main.async {
                    completion(querySnapshot!.documents, err)
                }
            }
        }
    }
    
    //For getting a document in a collection
    //Will receive a closure
    func fetchDocumentInCollection(mainCollection: String, document: String, completion: @escaping (DocumentSnapshot?, Error?) -> Void) {
        db.collection(mainCollection)
            .document(document)
            .getDocument() { (snapshot, err) in
            if let err = err {
                print("Error getting documents: \(err)")
                DispatchQueue.main.async {
                    completion(snapshot, err)
                }
            } else {
                DispatchQueue.main.async {
                    completion(snapshot, err)
                }
            }
        }
    }
    
    //For getting a document in a second level collection
    //Will receive a closure
    func fetchDocumentDeeper(
        mainCollection: String,
        mainDocument: String,
        secondaryCollection: String,
        document: String,
        completion: @escaping (DocumentSnapshot?, Error?) -> Void){
        
        db.collection(mainCollection)
            .document(mainDocument)
            .collection(secondaryCollection)
            .document(document)
            .getDocument() { (snapshot, err) in
            if let err = err {
                print("Error getting documents: \(err)")
                DispatchQueue.main.async {
                    completion(snapshot, err)
                }
            } else {
                DispatchQueue.main.async {
                    completion(snapshot, err)
                }
            }
        }
    }
  
    //For getting a document in a second level collection
    //Will receive a closure
    func fetchDocument(
        mainCollection: String,
        document: String,
        completion: @escaping (DocumentSnapshot?, Error?) -> Void){
        
        db.collection(mainCollection)
            .document(document)
            .getDocument() { (snapshot, err) in
            if let err = err {
                print("Error getting documents: \(err)")
                DispatchQueue.main.async {
                    completion(snapshot, err)
                }
            } else {
                DispatchQueue.main.async {
                    completion(snapshot, err)
                }
            }
        }
    }
}
